# pemmican: notifies users of Raspberry Pi 5 power issues
#
# Copyright (c) 2024 Dave Jones <dave.jones@canonical.com>
# Copyright (c) 2024 Canonical Ltd.
#
# SPDX-License-Identifier: GPL-3.0

import struct

from pathlib import Path


DT_POWER = Path('/proc/device-tree/chosen/power')


def reset_brownout():
    """
    Returns :data:`True` if the device-tree reports that a power brownout
    (undervolt condition) was the cause of the last reset. Raises
    :exc:`OSError` if the reset condition cannot be queried (e.g. if this is
    executed on a non-Raspberry Pi).
    """
    # D-T values are big-endian (hence the > prefix)
    fmt = struct.Struct('>I')
    with (DT_POWER / 'power_reset').open('rb') as f:
        value, = fmt.unpack(f.read(fmt.size))
        return bool(value & 0x02)


def psu_max_current():
    """
    Returns the maximum current negotiated with the PSU by the power supply in
    mA. Ideally this should be 5000 (indicating a power supply capable of 5V at
    5A), but may be 3000 or lower. Raises :exc:`OSError` if the maximum current
    could not be queried (e.g. if this is executed on a non-Raspberry Pi).
    """
    fmt = struct.Struct('>I')
    with (DT_POWER / 'max_current').open('rb') as f:
        value, = fmt.unpack(f.read(fmt.size))
        return value



# pemmican: notifies users of Raspberry Pi 5 power issues
#
# Copyright (c) 2024 Dave Jones <dave.jones@canonical.com>
# Copyright (c) 2024 Canonical Ltd.
#
# SPDX-License-Identifier: GPL-3.0

import pytest
from unittest import mock

from pemmican.notify import Notifications


def test_notifications_init(dbus):
    notifier = Notifications()
    assert notifier.on_closed is None
    assert notifier.on_action is None
    assert notifier.pending == frozenset()
    assert dbus.SessionBus().add_signal_receiver.call_count == 2


def test_notifications_init_with_bus(dbus):
    my_bus = dbus.SessionBus()
    notifier = Notifications(my_bus)
    assert notifier.on_closed is None
    assert notifier.on_action is None
    assert notifier.pending == frozenset()
    assert my_bus.add_signal_receiver.call_count == 2


def test_notifications_info(notify_intf):
    notifier = Notifications()

    notify_intf.GetServerInformation.return_value = (
        'foo-shell', 'FOO', '1.0', '1.2')
    assert notifier.get_server_info() == {
        'name': 'foo-shell',
        'vendor': 'FOO',
        'version': '1.0',
        'spec_version': '1.2',
    }


def test_notifications_caps(notify_intf):
    notifier = Notifications()

    caps = ['persistence', 'body', 'body-hyperlink']
    notify_intf.GetCapabilities.return_value = caps
    assert notifier.get_capabilities() == caps


def test_notifications_notify(notify_intf):
    notifier = Notifications()
    notify_intf.Notify.return_value = 9

    assert notifier.notify('Foo Info', body='The foo has started') == 9
    assert notifier.pending == {9}
    assert notify_intf.Notify.call_args.args == (
        '', 0, '', 'Foo Info', 'The foo has started', [], {}, -1)
    assert notify_intf.Notify.call_args.kwargs == {}

    notify_intf.Notify.return_value = 10
    assert notifier.notify(
        'Foo Error', body='The foo is on fire!',
        actions=[('put-it-out', 'Extinguish!'), ('let-it-burn', 'Leave It')],
        hints={'urgency': 2}) == 10
    assert notifier.pending == {9, 10}
    assert notify_intf.Notify.call_args.args == (
        '', 0, '', 'Foo Error', 'The foo is on fire!',
        ['put-it-out', 'Extinguish!', 'let-it-burn', 'Leave It'],
        {'urgency': 2}, -1)
    assert notify_intf.Notify.call_args.kwargs == {}


def test_notifications_not_pending(notify_intf):
    notifier = Notifications()
    notify_intf.Notify.return_value = 0

    assert notifier.notify('Foo Info', body='The foo has started') == 0
    # Ensure we don't set 0 (which is not a valid message ID) as a pending
    # identity
    assert notifier.pending == set()


def test_notifications_remove(notify_intf):
    notifier = Notifications()
    notifier.remove(10)
    assert notify_intf.CloseNotification.call_args.args == (10,)
    assert notify_intf.CloseNotification.call_args.kwargs == {}


def test_notifications_action(notify_intf):
    notifier = Notifications()
    notifier.on_action = mock.Mock()

    notifier._action_invoked(9, 'put-it-out')
    assert notifier.on_action.call_count == 0

    notify_intf.Notify.return_value = 10
    assert notifier.notify(
        'Foo Error', body='The foo is on fire!',
        actions=[('put-it-out', 'Extinguish!'), ('let-it-burn', 'Leave It')],
        hints={'urgency': 2}) == 10
    notifier._action_invoked(10, 'put-it-out')
    assert notifier.on_action.call_count == 1
    assert notifier.on_action.call_args.args == (10, 'put-it-out')
    notifier._action_invoked(11, 'put-it-out')
    assert notifier.on_action.call_count == 1


def test_notifications_no_action(notify_intf):
    notifier = Notifications()
    notifier.on_action = None

    notify_intf.Notify.return_value = 11
    assert notifier.notify(
        'Foo Error', body='The foo is on fire!',
        actions=[('put-it-out', 'Extinguish!'), ('let-it-burn', 'Leave It')],
        hints={'urgency': 2}) == 11
    # Ensure if on_action is None, the _action_invoked handler does not raise
    # an exception, even if the ID is pending and everything else is valid
    notifier._action_invoked(11, 'put-it-out')


def test_notifications_closed(notify_intf):
    notifier = Notifications()
    notifier.on_closed = mock.Mock()

    notifier._notification_closed(9, 2)
    assert notifier.on_closed.call_count == 0

    notify_intf.Notify.return_value = 10
    assert notifier.notify(
        'Foo Error', body='The foo is on fire!',
        actions=[('put-it-out', 'Extinguish!'), ('let-it-burn', 'Leave It')],
        hints={'urgency': 2}) == 10
    notifier._notification_closed(10, 2)
    assert notifier.on_closed.call_count == 1
    assert notifier.on_closed.call_args.args == (10, 2)
    notifier._notification_closed(11, 2)
    assert notifier.on_closed.call_count == 1


def test_notifications_closed_but_none(notify_intf):
    notifier = Notifications()
    notifier.on_closed = None

    notify_intf.Notify.return_value = 11
    assert notifier.notify(
        'Foo Error', body='The foo is on fire!',
        actions=[('put-it-out', 'Extinguish!'), ('let-it-burn', 'Leave It')],
        hints={'urgency': 2}) == 11
    # Ensure if on_closed is None, the _notification_closed handler does not
    # raise an exception, even if the ID is pending and everything else is
    # valid
    notifier._notification_closed(11, 2)



# pemmican: notifies users of Raspberry Pi 5 power issues
#
# Copyright (c) 2024 Dave Jones <dave.jones@canonical.com>
# Copyright (c) 2024 Canonical Ltd.
#
# SPDX-License-Identifier: GPL-3.0

import dbus


class Notifications:
    """
    Wraps the methods of the DBus Notifications freedesktop service. If *bus*
    is not specified, it defaults to the DBus :class:`~dbus.SessionBus`. As
    the instance needs to receive signals, the bus must be associated with a
    "main loop", e.g. :class:`DBusGMainLoop`.
    """
    SERVICE_NAME = 'org.freedesktop.Notifications'
    OBJECT_PATH = '/org/freedesktop/Notifications'

    def __init__(self, bus=None):
        if bus is None:
            bus = dbus.SessionBus()
        server = bus.get_object(self.SERVICE_NAME, self.OBJECT_PATH)
        self._intf = dbus.Interface(server, dbus_interface=self.SERVICE_NAME)
        self._pending = set()
        self.on_closed = None
        self.on_action = None
        bus.add_signal_receiver(
            self._notification_closed, 'NotificationClosed', self.SERVICE_NAME)
        bus.add_signal_receiver(
            self._action_invoked, 'ActionInvoked', self.SERVICE_NAME)

    def get_server_info(self):
        """
        Returns a :class:`dict` with keys "name", "vendor", "version",
        and "spec_version" mapped to the values returned by the service.
        """
        return {
            key: value
            for key, value in zip(
                ('name', 'vendor', 'version', 'spec_version'),
                self._intf.GetServerInformation()
            )
        }

    def get_capabilities(self):
        """
        Return a :class:`list` of :class:`str` detailing the capabilities of
        the notification service.

        Typical values include "actions", the server will display specified
        actions to the user, "body", the server supports notification bodies,
        "body-markup", the server supports limited markup in the notification
        body, and "persistence", the server supports persistence of
        notifications. See the `Desktop Notifications Specification`_ for all
        possible return values.

        .. _Desktop Notifications Specification:
            https://specifications.freedesktop.org/notification-spec/latest/
        """
        return self._intf.GetCapabilities()

    def notify(self, summary, *, body='', app_name='', app_icon='',
               replaces_id=0, actions=None, hints=None, timeout=-1):
        """
        Send a notification to the service.

        The *summary*, a brief :class:`str` summarizing the message, must be
        included. The *body* is the full message which may optionally contain
        a limited form of XML markup, if "body-markup" is included in the
        result of :meth:`get_capabilities`.

        The optional *app_name* and *app_icon* strings identify the application
        sending the notification.

        The *actions* parameter specifies a :class:`list` of ``(id, label)``
        tuples where *id* is the action identifier, and *label* is the string
        to display for the action. This is only supported if "actions" is
        included in the result of :meth:`get_capabilities`.

        The *hints* parameter is a :class:`dict` of optional hints passed to
        the notification service. A common hint is "urgency" which takes the
        value 0 (low), 1 (normal), or 2 (critical).

        The *timeout* specifies how long the notification will remain
        displayed. The default of -1 is system-dependent. 0 indicates a
        notification never times out.

        The return value is the ID of the displayed notification. This may be
        used in future calls with the *replaces_id* parameter to replace prior
        notifications.
        """
        if actions is None:
            actions = []
        else:
            actions = [item for (id, label) in actions for item in (id, label)]
        if hints is None:
            hints = {}
        msg_id = self._intf.Notify(
            app_name, replaces_id, app_icon, summary, body, actions, hints,
            timeout)
        if msg_id:
            self._pending.add(msg_id)
        return msg_id

    def remove(self, msg_id):
        """
        Forces the notification identified by *msg_id* (an :class:`int` as
        returned by :meth:`notify`) to be removed from display.
        """
        self._intf.CloseNotification(msg_id)

    @property
    def pending(self):
        """
        Returns the set of pending message identifiers, as returned by
        :meth:`notify`.
        """
        return frozenset(self._pending)

    def _action_invoked(self, msg_id, action_id):
        """
        If *msg_id* is a message we sent, call :attr:`on_action` with the
        identifier, and the *action_id* (one of the ids passed as the *actions*
        parameter to :meth:`notify`) that was invoked by the user.
        """
        if msg_id in self._pending:
            if self.on_action is not None:
                self.on_action(msg_id, action_id)

    def _notification_closed(self, msg_id, reason):
        """
        If *msg_id* is a message we sent, call :attr:`on_closed` with the
        identifier and the *reason* for closing. See the `Desktop Notifications
        Specification`_ for valid reason codes.

        .. _Desktop Notifications Specification:
            https://specifications.freedesktop.org/notification-spec/latest/
        """
        try:
            self._pending.remove(msg_id)
        except KeyError:
            pass
        else:
            if self.on_closed is not None:
                self.on_closed(msg_id, reason)



# pemmican: notifies users of Raspberry Pi 5 power issues
#
# Copyright (c) 2024 Dave Jones <dave.jones@canonical.com>
# Copyright (c) 2024 Canonical Ltd.
#
# SPDX-License-Identifier: GPL-3.0

from unittest import mock

import pytest

from pemmican.cli import *


def test_help(capsys):
    with pytest.raises(SystemExit) as err:
        main(['--version'])
    assert err.value.code == 0
    capture = capsys.readouterr()
    assert capture.out.strip() == '1.0.2'

    with pytest.raises(SystemExit) as err:
        main(['--help'])
    assert err.value.code == 0
    capture = capsys.readouterr()
    assert capture.out.strip().startswith('usage:')


def test_lang_fallback(capsys):
    import locale
    with mock.patch('pemmican.lang.locale.setlocale') as setlocale:
        setlocale.side_effect = [locale.Error, None]
        with pytest.raises(SystemExit) as err:
            main(['--version'])
        assert err.value.code == 0
        # Ensure we fell back to C locale on initial failure
        assert setlocale.call_count == 2
        assert setlocale.call_args.args == (locale.LC_ALL, 'C')


def test_regular_operation(capsys):
    with (
        mock.patch('pemmican.cli.reset_brownout') as reset_brownout,
        mock.patch('pemmican.cli.psu_max_current') as psu_max_current,
    ):
        reset_brownout.return_value = False
        psu_max_current.return_value = 5000
        assert main([]) == 0
        capture = capsys.readouterr()
        assert capture.out.strip() == ''


def test_non_pi_operation(capsys):
    with (
        mock.patch('pemmican.cli.reset_brownout') as reset_brownout,
        mock.patch('pemmican.cli.psu_max_current') as psu_max_current,
    ):
        reset_brownout.side_effect = FileNotFoundError('no such file or directory')
        psu_max_current.return_value = 5000
        assert main([]) == 0
        capture = capsys.readouterr()
        assert capture.out.strip() == ''


def test_brownout(capsys):
    with (
        mock.patch('pemmican.cli.reset_brownout') as reset_brownout,
        mock.patch('pemmican.cli.psu_max_current') as psu_max_current,
    ):
        reset_brownout.return_value = True
        psu_max_current.return_value = 3000
        assert main([]) == 0
        capture = capsys.readouterr()
        out = capture.out.strip()
        assert out.startswith('Reset due to low power')
        assert 'man:pemmican-cli(1)' in out
        assert RPI_PSU_URL in out
        # Note: PSU max current is also reported low (3A), but this doesn't
        # result in output
        assert '5A' not in out


def test_max_current(capsys):
    with (
        mock.patch('pemmican.cli.reset_brownout') as reset_brownout,
        mock.patch('pemmican.cli.psu_max_current') as psu_max_current,
    ):
        reset_brownout.return_value = False
        psu_max_current.return_value = 3000
        assert main([]) == 0
        capture = capsys.readouterr()
        out = capture.out.strip()
        assert out.startswith('This power supply is not capable')
        assert 'man:pemmican-cli(1)' in out
        assert RPI_PSU_URL in out
        assert 'Reset due to low power' not in out
